#include "llvm/ADT/APInt.h"
#include "llvm/IR/Verifier.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"
#include "llvm/ExecutionEngine/GenericValue.h"
#include "llvm/ExecutionEngine/MCJIT.h"
#include "llvm/IR/Argument.h"
#include "llvm/IR/BasicBlock.h"
#include "llvm/IR/Constants.h"
#include "llvm/IR/DerivedTypes.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/InstrTypes.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/Type.h"
#include "llvm/Support/Casting.h"
#include "llvm/Support/TargetSelect.h"
#include "llvm/Support/raw_ostream.h"
#include <llvm/ExecutionEngine/SectionMemoryManager.h>
#include <llvm/IR/IRBuilder.h>
#include <llvm/ExecutionEngine/Orc/CompileUtils.h>
#include <llvm/ExecutionEngine/Orc/IRCompileLayer.h>
#include <llvm/ExecutionEngine/Orc/LambdaResolver.h>
#include <llvm/ExecutionEngine/Orc/RTDyldObjectLinkingLayer.h>
#include <algorithm>
#include <cstdlib>
#include <memory>
#include <string>
#include <vector>
#include <iostream>

using namespace llvm;

int (*add_func) ( int, int );

static Function *createJitFunction(std::shared_ptr<Module> M, LLVMContext &context) {
  IRBuilder<> builder = IRBuilder<>(context);

  // create function with right signature
  Function *addF =
    cast<Function>(M->getOrInsertFunction("jit_func", Type::getInt32Ty(context),
                                          Type::getInt32Ty(context),Type::getInt32Ty(context)));
  // entry basic block
  BasicBlock *entry = BasicBlock::Create(context, "entry", addF);
  builder.SetInsertPoint ( entry ); 
  
  // get parameters 
  std::vector<Value*> args;
  for(auto& arg : addF->args())
          args.push_back(&arg);
  Value* v1 = args[0];
  Value* v2 = args[1];

  // create instructions
  Value* sum = builder.CreateAdd ( v1, v2, "sum" );
  builder.CreateRet ( sum );
  return addF;
}



using LinkLayer = orc::RTDyldObjectLinkingLayer;
using LinkLayer = orc::RTDyldObjectLinkingLayer;
using Compiler = orc::SimpleCompiler;
using CompileLayer = orc::IRCompileLayer<LinkLayer, Compiler>;

static JITSymbol dummy_lookup(const std::string& name) {
    return JITSymbol(nullptr);
}

int main() {

  // create and put our function into it.
  LLVMContext context;
  std::shared_ptr < Module > jit_module ( new Module("JIT Module", context) );
  createJitFunction ( jit_module, context );

  // print module
  std::string moduleIR;
  raw_string_ostream OS ( moduleIR );
  OS << *jit_module;
  OS.flush();
  std::cout << "We just constructed this LLVM module:" << std::endl;
  std::cout << "--------------------------------------------------\n" << moduleIR << std::endl;
  std::cout << "--------------------------------------------------\n" << std::endl;

  // verify moule
  if ( llvm::verifyModule(*jit_module, &llvm::outs()) ) {
      std::cout << "\nModule verification failed." << std::endl;
      return false;
  }
  
  // compile
  InitializeNativeTarget();
  InitializeNativeTargetAsmPrinter();
  std::shared_ptr < TargetMachine > target ( EngineBuilder().selectTarget() );
  std::shared_ptr < LinkLayer > link_layer = std::make_shared<LinkLayer>([]() { return std::make_shared<SectionMemoryManager>(); });
  std::shared_ptr < CompileLayer > compile_layer = std::make_shared<CompileLayer> ( *link_layer, Compiler(*target) );
  auto jit_module_handle = cantFail ( compile_layer->addModule( std::shared_ptr<Module> ( jit_module ), 
      orc::createLambdaResolver ( dummy_lookup, dummy_lookup ) ) );

  // get function pointer
  JITSymbol symbol = compile_layer->findSymbolIn( jit_module_handle, "jit_func", false);
  if ( !symbol ) {
      std::cout << "Error: JIT function not found" << std::endl;
      return false;
  }
  add_func = (decltype(add_func))cantFail(symbol.getAddress());

  // execute jit function
  std::cout << "Executing JIT function:" << std::endl;
  std::cout << "add 42 + 18 = " << add_func ( 42, 18 ) << std::endl;

}
